/*
 * Common values for GOSTHASH94 algorithms
 */

#ifndef _CRYPTO_GOSTHASH94_H
#define _CRYPTO_GOSTHASH94_H

#include <linux/types.h>
#include <linux/crypto.h>

#define GOSTHASH94_DIGEST_SIZE	32
#define GOSTHASH94_BLOCK_SIZE	32

#endif
